from string import Template

import numpy as np

CELL_TEMPLATE = Template(
"""Ndim=$ndim
cellv1= $cellv1
cellv2= $cellv2
cellv3= $cellv3
"""
)

MWFN_TEMPLATE = Template(
    """# Generate by Thunder-ASE
Wfntype=$wfntype
Charge=$charge
Naelec=$naelec
Nbelec=$nbelec
E_tot=$e_tot
VT_ratio=$vt_ratio
$cell_info
# Atom information
Ncenter=$ncenter
$$Centers
$atoms_coord

# Basis function information
Nbasis=    $nbasis
Nindbasis= $nindbasis
Nprims=    $nprims
Nshell=    $nshell
Nprimshell=$nprimshell
$$Shell types
$shell_types
$$Shell centers
$shell_centers
$$Shell contraction degrees
$shell_contraction_degress
$$Primitive exponents
$primitive_exponents
$$Contraction coefficients
$contraction_coefficients

# Orbital information (nindbasis orbitals)

$orbital_coeffs
"""
)

MWFN_DEFAULT = {
    'wfntype': 0,
    'charge': 0.0,
    'naelec': None,
    'nbelec': None,
    'e_tot': 0.0,
    'vt_ratio': 0.0,
    'ndim': None,
    'cellv1': None,
    'cellv2': None,
    'cellv3': None,
    'ncenter': None,
    'atoms_coord': None,
    'nbasis': None,
    'nindbasis': None,
    'nprims': None,
    'nshell': None,
    'nprimshell': None,
    'shell_types': None,
    'shell_centers': None,
    'shell_contraction_degress': None,
    'primitive_exponents': None,
    'contraction_coefficients': None,
}

MWFN_FORMAT = {
    'wfntype': '{:4d}',
    'charge': '{:15.6f}',
    'naelec': '{:15.6f}',
    'nbelec': '{:15.6f}',
    'e_tot': '{:16.8E}',
    'vt_ratio': '{:12.8f}',
    'ndim': '{:4d}',
    'cellv1': '{:12.8f}',
    'cellv2': '{:12.8f}',
    'cellv3': '{:12.8f}',
    'ncenter': '{:8d}',
    'atoms_coord': '{:6d} {:<2s}{:4d}{:6.1f}{:16.8f}{:16.8f}{:16.8f}',
    'nbasis': '{:8d}',
    'nindbasis': '{:8d}',
    'nprims': '{:8d}',
    'nshell': '{:8d}',
    'nprimshell': '{:8d}',
    'shell_types': '{:3d}',
    'shell_centers': '{:8d}',
    'shell_contraction_degress': '{:4d}',
    'primitive_exponents': '{:16.8E}',
    'contraction_coefficients': '{:16.8E}',
    'MO_coefficients': '{:16.8E}',
}

KEY_LINE_LEN = {
    'shell_types': 25,
    'shell_centers': 10,
    'shell_contraction_degress': 20,
    'primitive_exponents': 5,
    'contraction_coefficients': 5,
    'MO_coefficients': 5,
}


def format_data(key, data):
    maxlen = KEY_LINE_LEN[key]
    res = len(data) % maxlen
    nline = int(len(data) / maxlen)
    result = [(MWFN_FORMAT[key] * maxlen).format(*data[i*maxlen:(i+1)*maxlen])
              for i in range(nline)]
    result.append((MWFN_FORMAT[key] * res).format(*data[-res:]))
    result = [i for i in result if len(i.strip()) > 0]
    return '\n'.join(result)


def read_cdcoeffs(filename):
    with open(filename, 'r') as f:
        lines = f.readlines()

    block_start, block_end = [], []
    for il, line in enumerate(lines):
        if "Kpoint=" in line:
            block_start.append(il+1)
            if il > 0:
                block_end.append(il-1)
    block_end.append(len(lines))

    coeffs_kpoints = []
    for start, end in zip(block_start, block_end):
        coeff_orbitals = []
        coeff_orbital = {
            'info': [],
            'coeff': []
        }
        for line in lines[start:end]:
            content = line.strip()
            if len(content) == 0:
                if len(coeff_orbital['info']) > 0:
                    coeff_orbitals.append(coeff_orbital)
                coeff_orbital = {
                    'info': [],
                    'coeff': []
                }
                continue
            if 'Index' in content:
                start_orbital = True
            elif '$Coeff' in content:
                coeff_orbital['info'].append(line)
                start_orbital = False
                continue
            if start_orbital:
                coeff_orbital['info'].append(line)
            else:
                coeff_orbital['coeff'] += [float(i) for i in content.split()]

        coeffs_kpoints.append(coeff_orbitals)
    return coeffs_kpoints


def reorder_cdcoeffs(coeff_list, shell_types):
    """
    Re-order the orbital from (Y, Z, X) and (D-2, D-1, D0, D+1, D+2) to
    (X, Y, Z) and (D0, D+1, D-1, D+2, D-2)

    :param coeff_list:
    :param shell_types:
    :return:
    """
    new_coeff_list = []
    idx_curr = 0
    for st in shell_types:
        if st == 0:  # s orbital
            new_coeff_list += [coeff_list[idx_curr]]
        elif st == 1:  # p orbital
            new_coeff_list += [
                coeff_list[idx_curr+2],
                coeff_list[idx_curr],
                coeff_list[idx_curr+1]
            ]
        elif st == -2:  # pure d orbital. The scale factor, see fireball/b.FUNCTIONS/clm.f90 and Ylm.f90
            new_coeff_list += [
                coeff_list[idx_curr+2],  # D0
                coeff_list[idx_curr+3],  # D+1
                coeff_list[idx_curr+1],  # D-1
                coeff_list[idx_curr+4],  # D+2
                coeff_list[idx_curr+0],  # D-2
            ]
        else:
            NotImplementedError

        idx_curr += np.abs(st) * 2 + 1

    return new_coeff_list



import argparse
import os.path

import numpy as np
from scipy.optimize import minimize, basinhopping
import matplotlib.pyplot as plt

from thunder_ase.utils import ordinal
from thunder_ase.utils.shell_dict import SHELL_NUM, SHELL_NAME
from ase.data import chemical_symbols
from ase.units import Bohr


# read wf file
def read_wf(filename):
    with open(filename, 'r') as f:
        lines = f.readlines()
    data = [list(map(float, line.strip().split())) for line in lines if len(line.strip().split()) == 2]
    return np.asarray(data)


def read_info(filename='Fdata/info.dat'):
    """
    :param filename:
    :return: dict, {'H': {
        'number': 1,
        'nshell': 1,
        'shells': [0, 1],
        'occupation': [1.0, 0.0],
    }}
    """
    with open(filename, 'r') as f:
        lines = f.readlines()

    result_dict = {}

    status = 0  # 0: start block, 1: in block, 2: need to end block
    for il, line in enumerate(lines):
        content = line.strip()
        if len(content) == 0:
            continue
        if content == '=' * 70:
            if status == 0:
                status = 1
            elif status == 2:
                status = 0
                continue

        if status == 1:
            result_dict[lines[il + 2].strip().split()[0]] = {
                'number': int(lines[il + 3].strip().split()[0]),
                'nshell': int(lines[il + 5].strip().split()[0]),
                'shells': [int(i) for i in lines[il + 6].strip().split()],
                'occupation': [float(i) for i in lines[il + 7].strip().split()],
            }
            status = 2
        else:
            continue

    return result_dict


def norm_c(a, shell=0):
    # normalization constant for pure Gaussian basis functions
    # see https://iodata.readthedocs.io/en/latest/basis.html
    norm = (2 * a / np.pi) ** 0.75  # s orbital
    if shell == 1:  # p orbital
        norm = norm * 2 * np.sqrt(a)
    elif shell == 2:  # d orbital
        norm = norm * 4 * a / np.sqrt(3)
    elif shell > 2:
        raise NotImplementedError
    return norm


def dnorm_c(a, shell=0):
    # derivative of constant part of normalization constant
    dnorm = (2 * a / np.pi) ** (-0.25) * (2 / np.pi)  # s orbital
    if shell == 1:  # p orbital
        dnorm = dnorm * 2 * np.sqrt(a) + norm_c(a, 0) * 2 / np.sqrt(a)
    elif shell == 2:  # d orbital
        dnorm = dnorm * 4 * a / np.sqrt(3) + norm_c(a, 0) * 4 / np.sqrt(3)
    elif shell > 2:
        raise NotImplementedError
    return dnorm


def gaussian(r, l=0, A=np.array([1.0]), a=np.array([0.1])):
    # gaussian function
    f = np.sum([Ai * (r ** l) * np.exp(-ai * (r ** 2)) * norm_c(ai, l)
                for Ai, ai in zip(A, a)], axis=0)
    return f


# loss function
def loss_function(x, *args):
    """
    TODO: use sum( (4 * pi * r**2 * (wf**2 - wf0**2))**2 ) as loss function
    :param x:
    :param args:
    :return:
    """
    len_x = int(len(x) / 2)
    alpha = 10 ** x[0:len_x]
    A = np.asarray(x[len_x:])
    l, r, Y = args
    if r.shape != Y.shape:
        raise ValueError('Shapes for args[1] and args[2] are not equal!')

    loss = np.mean((gaussian(r, l, A, alpha) - Y) ** 2)
    return loss


def loss_jac(x, *args):
    len_x = int(len(x) / 2)
    alpha = 10 ** x[0:len_x]
    A = np.asarray(x[len_x:])
    l, r, Y = args
    if r.shape != Y.shape:
        raise ValueError('Shapes for args[1] and args[2] are not equal!')
    d_alpha = np.asarray([Ai * (r ** l) * np.exp(-ai * (r ** 2)) * ai * np.log(10) *
                          (dnorm_c(ai, l) - r ** 2 * norm_c(ai, l)) for Ai, ai in zip(A, alpha)])
    d_A = np.asarray([(r ** l) * np.exp(-ai * (r ** 2)) * norm_c(ai, l) for ai in alpha])
    der = np.concatenate([d_alpha, d_A])
    jac = 2 * np.mean((gaussian(r, l, A, alpha) - Y) * der, axis=1)
    return jac


def fit_wf(data, x0, l=0, bnds=None):
    R, Y = np.asarray(data)
    nz = int(len(x0) / 2)
    if bnds is not None:
        bnds = [bnds[0]] * nz + [bnds[1]] * nz
    res = minimize(loss_function, x0, bounds=bnds, args=(l, R, Y))
    alpha = res.x[0:nz]  # exponential parameters,
    Ae = res.x[nz:]  # coefficients
    error = res.fun
    return [Ae, alpha, error]


def fit_wf_from_random(data, l=0, tol=1e-5, Nzeta0=3, Nzeta_max=10, Niter=3, bnds=None):
    """

    :param Niter:
    :param data: shape = [N, 2]
    :param l: principal quantum number
    :param tol: Error tolerance
    :param Nzeta: number of gaussian
    :param bnds: boundary for fitting
    :return:
    """
    R, Y = np.asarray(data)
    error = np.inf
    result_res = None

    print("Fitting parameters: shell = {}, Nzeta0 = {}, Nzeta_max = {}, Ntry = {}, tol = {}"
          .format(l, Nzeta0, Nzeta_max, Niter, tol))

    for nz in range(Nzeta0, Nzeta_max + 1):
        if bnds is None:
            bnds_lb = [-1.5] * nz + [-1.0] * nz
            bnds_ub = [4.0] * nz + [1.0] * nz
            bounds = list(zip(bnds_lb, bnds_ub))
        else:
            bounds = bnds
        for itry in range(Niter):
            print("::: The {} try for Nzeta = {} ...".format(ordinal(itry + 1), nz))
            init_alpha = np.random.random(nz) * 3.5 - 1.0  # -1.0 ~ 2.5
            init_coeff = np.random.random(nz) * 0.5 + 0.5  # 0.5 ~ 1.0
            init_guess = np.concatenate([init_alpha, init_coeff])
            res = basinhopping(loss_function, init_guess, T=0.0001, niter=100, disp=False, stepsize=0.5,
                               niter_success=50,
                               interval=20,
                               stepwise_factor=0.8,
                               minimizer_kwargs={'args': (l, R, Y),
                                                 'bounds': bounds,
                                                 'jac': loss_jac,
                                                 }
                               )
            if res.fun < error:
                result_res = res
                error = res.fun
                if error < tol:
                    alpha = 10 ** result_res.x[0:nz]  # exponential parameters,
                    Ae = result_res.x[nz:]  # coefficients
                    print("Success!: Fitting error {} meet the tolerance {} for {} gaussians."
                          .format(error, tol, nz))
                    return [Ae, alpha, error]
        print("Fitting error {} didn't meet the tolerance {} for {} gaussians after {} try."
              .format(error, tol, nz, Niter))

    alpha = 10 ** result_res.x[0:Nzeta_max]  # exponential parameters,
    Ae = result_res.x[Nzeta_max:]  # coefficients
    print("Warning: Fitting error {} didn't meet the tolerance {} for {} gaussians."
          .format(error, tol, Nzeta_max))

    return [Ae, alpha, error]


def plot_fitting(data, l, Ae, ae, output=None):
    R, Y = np.asarray(data)
    plt.plot(R, Y, '*')
    plt.plot(R, gaussian(R, l, Ae, ae))
    Y_min = min(Y)
    Y_max = max(Y)
    Y_range = Y_max - Y_min
    ymin = Y_min - Y_range * 0.2
    ymax = Y_max + Y_range * 0.2
    plt.ylim([ymin, ymax])
    for Ai, ai in zip(Ae, ae):
        plt.plot(R, gaussian(R, l, [Ai], [ai]))
    if output is not None:
        plt.savefig(output)
    else:
        plt.show()


def get_initial_gbs_guess(element_num, primary_num, shell_name):
    # loop GBS, find minimum element contain excited_pn and shell_name
    pn = str(primary_num)
    for n_element in chemical_symbols[element_num:]:
        if n_element not in GBS:
            continue
        if pn not in GBS[n_element]:
            continue
        if shell_name in GBS[n_element][pn]:
            gauss = GBS[n_element][pn][shell_name]
            if n_element != chemical_symbols[element_num]:
                print("Initial guess for {} {}{} is from {}!".format(
                    chemical_symbols[element_num], pn, shell_name, n_element))
            return np.concatenate(np.asarray(gauss).T)

    print("No initial guess for {} {}{}!".format(chemical_symbols[element_num], pn, shell_name))
    print("Maybe too large element number!")
    raise NotImplementedError


def expand_data(wf_data):
    """
    Expand the cutoff to current_cutoff + 3.0 angstrom.
    This is to get better fitting for the gaussian tail.
    :param wf_data:
    :return:
    """
    r, y = wf_data
    r_max = r.max()
    dr = r_max / len(r)
    r_plus = 3.0
    r_extend = np.arange(r_max + dr, r_max + r_plus, dr)
    new_r = np.concatenate([r, r_extend])
    new_y = np.concatenate([y, np.zeros(len(r_extend))])

    return np.asarray([new_r, new_y])


# run this after begin.x
def fit_gaussian(prog='fit_gaussians',
                 description='Fit Fireball basis set to Gaussian-type basis set.'):
    parser = argparse.ArgumentParser(
        prog=prog,
        description=description, )
    parser.add_argument('input_name', nargs='+', help='Fireball wave function file.')
    parser.add_argument('-p', '--plot', action='store_true')
    parser.add_argument('-t', '--tolerance', type=float, dest='tolerance', default=1e-4)
    parser.add_argument('-Nz', '--Nzeta_max', type=int, dest='nzeta_max', default=10)
    parser.add_argument('-Nz0', '--Nzeta0', type=int, dest='nzeta0', default=4)
    parser.add_argument('-Nt', '--Ntry', type=int, dest='ntry', default=3)
    args = parser.parse_args()

    for input_name in args.input_name:
        name_list = input_name.split('.')  # format: '001.wf-s0.dat', element_number, shell, is_excited
        element_number = int(name_list[0])
        shell_name, is_excited = name_list[1][-2:]
        shell = SHELL_NUM[shell_name]
        if not os.path.exists(input_name):
            print("{} doesn't exist!".format(input_name))
            raise FileNotFoundError
        wf_data = read_wf(input_name).T

        coeff_angular = Bohr**1.5
        # normalize wf_data for different l: sqrt((2*l+1) / (4*pi))
        if shell == 0:
            coeff_angular = coeff_angular * np.sqrt(1.0 / (4.0 * np.pi))
        elif shell == 1:
            coeff_angular = coeff_angular * np.sqrt(3.0 / (4.0 * np.pi))
        elif shell == 2:
            coeff_angular = coeff_angular * np.sqrt(5.0 / (4.0 * np.pi))
        else:
            NotImplementedError

        wf_data[1] = wf_data[1] * coeff_angular
        wf_data = expand_data(wf_data)
        # fitting from random
        Ae, ae, error = fit_wf_from_random(data=wf_data, l=shell,
                                           tol=args.tolerance,
                                           Nzeta_max=args.nzeta_max,
                                           Nzeta0=args.nzeta0,
                                           Niter=args.ntry)

        if args.plot:
            output_fig = "{:03d}.wf-{}{}-fitting.png".format(element_number, shell_name, is_excited)
            plot_fitting(wf_data, shell, Ae, ae, output=output_fig)

        output = "{:03d}.wf-{}{}.gbs".format(element_number, shell_name, is_excited)
        with open(output, 'w') as f:
            for A, a in zip(Ae, ae):
                f.write("{: .8E}  {: .8E}\n".format(a, A))  # order is alpha, co_effi


def read_gaussian(element_number, shell, is_excited, Fdata_path=None):
    shell_name = SHELL_NAME[shell]
    filename = os.path.join(Fdata_path, 'basis', "{:03d}.wf-{}{}.gbs".format(element_number, shell_name, is_excited))
    with open(filename, 'r') as f:
        lines = f.readlines()
    array = np.asarray([[float(i) for i in line.split()] for line in lines])
    return {'alpha': array[:, 0], 'coefficient': array[:, 1], 'degree': len(array)}



def ordinal(n: int):
    """
    Convert an integer to "1st", "2nd", etc,
    :param n: int
    :return: str
    """
    if 11 <= (n % 100) <= 13:
        suffix = 'th'
    else:
        suffix = ['th', 'st', 'nd', 'rd', 'th'][min(n % 10, 4)]
    return str(n) + suffix


from thunder_ase.utils.basis_set import fit_gaussian



"""
The interface of ASE for FIREBALL.
* create input files for FIREBALL
* run fireball calculations
* read output
"""
import os
import subprocess
from typing import Dict, Any
import ase
from ase.units import Hartree
import numpy as np
from ase.calculators.calculator import Calculator, CalculationFailed, all_changes
from ase.dft.kpoints import monkhorst_pack, kpoint_convert
from ase.geometry import get_distances
import ase.spacegroup
from ase.io import jsonio
from ase.dft.kpoints import BandPath
from thunder_ase.utils.mwfn import MWFN_FORMAT, MWFN_DEFAULT, MWFN_TEMPLATE, \
    CELL_TEMPLATE, format_data, read_cdcoeffs, reorder_cdcoeffs
from thunder_ase.utils.basis_set import read_info, read_gaussian
from thunder_ase.utils.shell_dict import SHELL_PRIMARY_NUMS, SHELL_PRIMITIVE


def get_kpts(atoms, size=None, offset=None, reduced=False, **kwargs):
    """
    Get reduced kpoints.
    Reference:
    * https://wiki.fysik.dtu.dk/ase/ase/dft/kpoints.html
    * https://wiki.fysik.dtu.dk/ase/ase/spacegroup/spacegroup.html
    :param reduced:
    :param offset:
    :param atoms:
    :param size:
    :return: np.array with shape (N, 4), coordinates and weights

    TODO: Need check the algorithm from gpaw/kpt_refine.py
    TODO: compare with vasp in several systems
    """
    # generate MP kpoints
    kpoints = monkhorst_pack(size) + np.asarray(offset)
    kpoints = kpoint_convert(cell_cv=atoms.cell, skpts_kc=kpoints)  # convert from scaled and cartesian coordinates
    nkpt = len(kpoints)
    if not reduced or nkpt <= 1:
        return [[kx, ky, kz, 1.0 / nkpt] for kx, ky, kz in kpoints]

    if 'symprec' in kwargs:
        symprec = kwargs['symprec']
    else:
        symprec = 1e-05  # for Cartesian coordinate in real space
    try:
        # get spacegroup from atoms ase.spacegroup.get_spacegroup
        sg = ase.spacegroup.get_spacegroup(atoms, symprec=symprec)
    except RuntimeError:
        print("Didn't find symmetry. No reducing is needed.")
        return [[kx, ky, kz, 1.0 / nkpt] for kx, ky, kz in kpoints]

    kpts_extend = []  # all the symmetry points for each kpts
    for kpt in kpoints:
        sites, kinds = sg.equivalent_sites([kpt])
        kpts_extend.append(sites)

    irreducible_dct = dict()
    kpoints_lst = kpoints.tolist()
    while len(kpoints_lst) > 0:
        kpt = kpoints_lst.pop(0)
        kpt_extend = kpts_extend.pop(0)
        key = tuple(kpt)
        irreducible_dct[key] = [kpt]
        # calc the distance between kpt_extend and kpts_extend
        if len(kpoints_lst) > 0:
            reduced_idx = []
            for idx, other_extend in enumerate(kpts_extend[:]):
                dist_array, dist = get_distances(kpt_extend, other_extend, cell=atoms.cell, pbc=True)
                if np.any(dist < symprec):
                    reduced_idx.append(idx)
            irreducible_dct[key] += [kpoints_lst[idx] for idx in reduced_idx]
            # update kpoints_lst and kpts_extend by removing the reduced_idx
            kpoints_lst = [v for idx, v in enumerate(kpoints_lst) if idx not in reduced_idx]
            kpts_extend = [v for idx, v in enumerate(kpts_extend) if idx not in reduced_idx]

    kpoints_weight = [[k[0], k[1], k[2], len(v) / nkpt] for k, v in irreducible_dct.items()]
    return kpoints_weight


options_params = {
    'nstepi': {'type': (int,), 'name': 'nstepi', 'default': 1},
    'nstepf': {'type': (int,), 'name': 'nstepf', 'default': 1},
    'qstate': {'type': (int,), 'name': 'qstate', 'default': 0},  # Extra electron. +1 mean adding 1 electron.
    'iquench': {'type': (int,), 'name': 'iquench', 'default': 0},
    't_initial': {'type': (int, float), 'name': 'T_initial', 'default': 300.0},
    't_final': {'type': (int, float), 'name': 'T_final', 'default': 0.0},
    't_want': {'type': (int, float), 'name': 'T_want', 'default': 300.0},
    'taurelax': {'type': (int, float), 'name': 'taurelax', 'default': 5.0},
    'efermi_t': {'type': (int, float), 'name': 'efermi_T', 'default': 100.0},
    'dt': {'type': (int, float), 'name': 'dt', 'default': 0.25},  # fs
    'iensemble': {'type': (int,), 'name': 'iensemble', 'default': 0},
    'iconstraint_rcm': {'type': (int,), 'name': 'iconstraint_rcm', 'default': 1},  # shift molecule to COM or not
    'iconstraint_vcm': {'type': (int,), 'name': 'iconstraint_vcm', 'default': 1},  # whether keep COM fixed during md
    'iconstraint_l': {'type': (int,), 'name': 'iconstraint_L', 'default': 0},
    'iconstraint_ke': {'type': (int,), 'name': 'iconstraint_KE', 'default': 1},
    'ifix_neighbors': {'type': (int,), 'name': 'ifix_neighbors', 'default': 0},
    'ifix_charges': {'type': (int,), 'name': 'ifix_CHARGES', 'default': 1},
    'max_scf_iterations_set': {'type': (int,), 'name': 'max_scf_iterations_set', 'default': 50},
    'scf_tolerance_set': {'type': (int, float), 'name': 'scf_tolerance_set', 'default': 0.00000001},
    'beta_set': {'type': (int, float), 'name': 'beta_set', 'default': 0.08},  # mix factor
    'ecut_set': {'type': (int, float), 'name': 'Ecut_set', 'default': 200.0},  # control mesh grid number
}

output_params = {
    'iwriteout_me_sandh': {'type': (int,), 'name': 'iwriteout_ME_SandH', 'default': 0},
    'iwriteout_density': {'type': (int,), 'name': 'iwriteout_density', 'default': 0},
    'iwriteout_cdcoeffs': {'type': (int,), 'name': 'iwriteout_cdcoeffs', 'default': 0},
    'iwriteout_charges': {'type': (int,), 'name': 'iwriteout_charges', 'default': 0},
    'iwriteout_energies': {'type': (int,), 'name': 'iwriteout_energies', 'default': 0},
    'iwriteout_populations': {'type': (int,), 'name': 'iwriteout_populations', 'default': 0},
    'iwriteout_forces': {'type': (int,), 'name': 'iwriteout_forces', 'default': 0},
    'iwriteout_neighbors': {'type': (int,), 'name': 'iwriteout_neighbors', 'default': 0},
    'iwriteout_dos': {'type': (int,), 'name': 'iwriteout_dos', 'default': 0},
    'iwriteout_abs': {'type': (int,), 'name': 'iwriteout_abs', 'default': 0},
    'iwriteout_ewf': {'type': (int,), 'name': 'iwriteout_ewf', 'default': 0},
}

xsfoptions_params = {
    'rho_surface_min': {'type': (int, float), 'name': 'rho_surface_min', 'default': 0.0005},
    'rho_surface_max': {'type': (int, float), 'name': 'rho_surface_max', 'default': 0.1},
}

calc_params = {
    'kpt_size': {'type': (list, np.array), 'name': 'kpt_size', 'default': None},
    'kpt_offset': {'type': (list, np.array), 'name': 'kpt_offset', 'default': None},
    'kpt_interval': {'type': (list, np.array, float, int), 'name': 'kpt_interval', 'default': None},
    'kpt_path': {'type': (BandPath, str, list, np.array), 'name': 'kpt_path', 'default': None},
    'nkpt': {'type': (int,), 'name': 'nkpt', 'default': None},  # n kpoints on path. Used if kpt_path is string.
    'kpt_reduced': {'type': (bool,), 'name': 'kpt_reduced', 'default': False},
}

fireball_params = options_params | output_params | xsfoptions_params | calc_params


def get_params_from_string(s):
    k, v = s.split('=')
    k = k.strip().lower()
    v = v.strip()
    if k not in fireball_params:
        raise KeyError
    vtype = fireball_params[k]['type']
    if float in vtype:
        v = float(v)
    else:
        v = vtype[0](v)
    return k, v


def write_params(dct, f):
    for k, v in dct.items():
        kname = fireball_params[k]['name']
        f.write("{} = {}\n".format(kname, v))


class GenerateFireballInput:
    def __init__(self, atoms=None, **kwargs):
        if atoms is not None:
            self._atoms = atoms.copy()
        else:
            self._atoms = None
        self.sname = '001'
        self.output_params = {}
        self.options_params = {}
        self.xsfoptions_params = {}

        self.kpt_size = fireball_params['kpt_size']['default']
        self.kpt_interval = fireball_params['kpt_interval']['default']
        self.kpt_offset = fireball_params['kpt_offset']['default']
        self.kpt_path = fireball_params['kpt_path']['default']
        self.nkpt = fireball_params['nkpt']['default']
        self.kpt_reduced = fireball_params['kpt_reduced']['default']
        self._kpoints = None

        self.check_kwargs(kwargs)

    @property
    def atoms(self):
        return self._atoms

    @atoms.setter
    def atoms(self, atoms):
        self._atoms = atoms.copy()

    def get_kpoints(self, reduced=True, **kwargs):
        if self._kpoints is not None:
            return self._kpoints

        if not np.all(self.atoms.pbc):
            return None

        if self.kpt_path is not None:
            if type(self.kpt_path) in (str,):
                lat = self.atoms.cell.get_bravais_lattice()
                path = lat.bandpath(self.kpt_path, npoints=self.nkpt)
                self.kpt_path = path
                kpts = path.kpts
            elif type(self.kpt_path) in (np.array, list, tuple):
                kpts = self.kpt_path
            elif type(self.kpt_path) in (BandPath,):  # TODO: kpt_path must be BandPath
                kpts = self.kpt_path.kpts
            else:
                raise TypeError

            kpts = kpoint_convert(cell_cv=self.atoms.cell, skpts_kc=kpts)  # convert to Cartesian coordinate
            self.nkpt = len(kpts)
            self._kpoints = [[k[0], k[1], k[2], 1.0 / self.nkpt] for k in kpts]
            return self._kpoints

        if self.kpt_offset is None:
            offset = [0., 0., 0.]
        else:
            offset = self.kpt_offset
        if self.kpt_size is None:
            if self.kpt_interval is None:
                size = [1, 1, 1]
            else:
                size = np.ceil(1 / self.atoms.cell.cellpar()[0:3] / self.kpt_interval)
        else:
            size = self.kpt_size

        self._kpoints = get_kpts(self.atoms, size=size, offset=offset, reduced=reduced, **kwargs)
        self.nkpt = len(self._kpoints)
        return self._kpoints

    def get_k_point_weights(self):
        return np.asarray(self.get_kpoints())[:, -1]

    def set_kpoints(self, kpoints):
        self._kpoints = kpoints

    def write_Fdata_inp(self, atoms=None, Fdata_path=None):
        if atoms is None:
            atoms = self.atoms
        if Fdata_path is None:
            Fdata_path = 'Fdata'
        species = sorted([i for i in set(atoms.numbers)])

        # TODO: if Fdata_path longer than 128, make a symbolic link.
        with open('Fdata.inp', 'w') as f:
            f.write("{}\n".format(len(species)))
            for s in species:
                f.write("{}\n".format(s))
            f.write("'{}'".format(Fdata_path))

    def check_kwargs(self, kwargs):
        for key, v in kwargs.items():
            k = key.lower()
            if k not in fireball_params:
                print("The option {} not supported!".format(k))
                raise KeyError
            if type(v) not in fireball_params[k]['type']:
                print("The type of {} should be {}".format(k, ' or '.join(fireball_params[k]['type'])))
                raise TypeError
            if k in output_params:
                self.output_params[k] = v
            elif k in options_params:
                self.options_params[k] = v
            elif k in xsfoptions_params:
                self.xsfoptions_params[k] = v
            elif k == 'kpt_size':
                self.kpt_size = v
            elif k == 'kpt_offset':
                self.kpt_offset = v
            elif k == 'kpt_interval':
                self.kpt_interval = v
            elif k == 'kpt_path':
                self.kpt_path = v
            elif k == 'nkpt':
                self.nkpt = v
            elif k == 'kpt_reduced':
                self.kpt_reduced = v
        return

    def write_options(self):
        with open('structures.inp', 'w') as f:
            f.write("1\n")
            f.write("{}.inp\n".format(self.sname))

            f.write("! Write out options - leave this line as comment line or null\n")

            f.write("&OUTPUT\n")
            write_params(self.output_params, f)
            f.write("&END\n")

            f.write("&OPTIONS\n")
            write_params(self.options_params, f)
            f.write("&END\n")

            f.write("&XSFOPTIONS\n")
            write_params(self.xsfoptions_params, f)
            f.write("&END\n")

    def write_atoms(self, atoms=None, pbc=None):
        """
        TODO: for non-PBC system, set cell to [0.000] * 9
        :param atoms:
        :param pbc:
        :return:
        """
        if atoms is not None:
            self.atoms = atoms.copy()

        with open(self.sname + ".inp", 'w') as f:
            if pbc is None:
                ipbc = 0 if np.any(self.atoms.pbc) else 1
            else:
                ipbc = 0 if np.any(pbc) else 1
            f.write("{:3d}{:12d}\n".format(len(self.atoms), ipbc))
            if ipbc == 0:
                cell = self.atoms.cell[:]
            else:
                cell = np.eye(3) * 999.0
            for row in cell:
                f.write("{:11.6f} {:11.6f} {:11.6f}\n".format(*row))
            for num, xyz in zip(self.atoms.numbers, self.atoms.positions):
                x, y, z = xyz
                f.write("{:3d} {:11.6f} {:11.6f} {:11.6f}\n".format(num, x, y, z))

    def write_kpts(self, reduced=True, **kwargs):
        """
        :param reduced:
        :param kwargs:
        :return:
        """
        with open(self.sname + '.KPOINTS', 'w') as f:
            f.write("{}\n".format(len(self.get_kpoints(reduced=reduced, **kwargs))))
            for k in self.get_kpoints(reduced=reduced, **kwargs):
                f.write("{:8.6f} {:8.6f} {:8.6f} {:8.6f}\n".format(*k))

    def write_input(self, atoms=None, Fdata_path=None):
        if atoms is not None:
            self.atoms = atoms.copy()
        self.write_Fdata_inp(atoms=self.atoms, Fdata_path=Fdata_path)
        self.write_options()
        self.write_atoms(pbc=self.atoms.pbc)
        if np.any(self.atoms.pbc):
            self.write_kpts(reduced=self.kpt_reduced)

    def read_options(self, input_file='structures.inp', read_atoms=False):
        # read structures.inp, get names for atoms and kpoints
        with open(input_file, 'r') as f:
            lines = f.readlines()

        natoms_list = int(lines[0].strip())
        sname_list = []
        for iline in lines[1:natoms_list + 1]:
            sname_list.append(iline.strip().strip('.inp'))

        loption, loutput, lxsfoptions = False, False, False
        for line in lines[natoms_list + 1:]:
            content = line.strip()
            if '!' in content:
                content = content.split('!')[0]
            if len(content) == 0:
                continue

            if '&OPTIONS' in content:
                loption = True
                continue
            if '&OUTPUT' in content:
                loutput = True
                continue
            if "&XSFOPTIONS" in content:
                lxsfoptions = True
                continue

            if loption:
                if '&END' in content:
                    loption = False
                    continue
                k, v = get_params_from_string(content)
                self.options_params[k] = v

            elif loutput:
                if '&END' in content:
                    loutput = False
                    continue
                k, v = get_params_from_string(content)
                self.output_params[k] = v
            elif lxsfoptions:
                if '&END' in content:
                    lxsfoptions = False
                    continue
                k, v = get_params_from_string(content)
                self.xsfoptions_params[k] = v

        if read_atoms:
            return sname_list

    def read_kpts(self, input_file=None):
        with open(input_file, 'r') as f:
            lines = f.readlines()
        nkpts = int(lines[0].strip())
        kpts = [list(map(float, line.strip().split())) for line in lines[1:] if len(line.strip()) > 0]
        if nkpts != len(kpts):
            raise ValueError("The kpoints number is not inconsistent.")
        self.set_kpoints(kpts)


class Fireball(GenerateFireballInput, Calculator):
    name = 'fireball'
    ase_objtype = 'fireball_calculator'  # For JSON storage

    # Environment commands
    env_commands = None

    implemented_properties = [
        'energy',
        'forces',
        'fermi',
    ]

    # Can be used later to set defaults
    default_parameters: Dict[str, Any] = {}

    def __init__(self, atoms=None,
                 Fdata_path='Fdata',
                 command='fireball.x',
                 **kwargs):
        self.__name__ = 'fireball'
        self._atoms = None
        self._eigenvalues = None
        self._shell_info = None
        GenerateFireballInput.__init__(self, **kwargs)  # TODO: set kwargs use set() function, see vasp calculator
        Calculator.__init__(self, atoms=atoms, **kwargs)
        if not os.path.isdir(Fdata_path):
            # check the existence of Fdata directory
            print("Error: Can't find Fdata! Pls Check the setting!")
            raise FileNotFoundError
        self.Fdata_path = Fdata_path
        self.command = command

    @property
    def atoms(self):
        return self._atoms

    @atoms.setter
    def atoms(self, atoms):
        if atoms is None:
            self._atoms = None
        else:
            self._atoms = atoms.copy()

    def calculate(self,
                  atoms=None,
                  properties=('energy',),
                  system_changes=tuple(all_changes)):
        """Do a fireball calculation in the specified directory.

        This will generate the necessary fireball input files, and then
        execute fireball. After execution, the energy, forces. etc. are read
        from the fireball output files.
        """

        if atoms is not None:
            self.atoms = atoms.copy()

        self.write_input(Fdata_path=self.Fdata_path)

        errorcode = self._run(command=self.command,
                              directory=self.directory)

        if errorcode:
            raise CalculationFailed(
                '{} in {} returned an error: {:d}'.format(
                    self.name, self.directory, errorcode))

        # self.update_atoms(atoms)
        self.read_results()

    def _run(self, command=None, directory=None):
        """Method to explicitly execute Fireball"""
        if command is None:
            command = self.command
        if directory is None:
            directory = self.directory
        errorcode = subprocess.call(command,
                                    shell=True,
                                    cwd=directory)
        return errorcode

    def read_results(self):
        output = self.sname + ".json"
        result = jsonio.read_json(output)
        if 'charges' in result['fireball'][-1]:
            shell_charges = result['fireball'][-1]['charges']
            result['fireball'][-1]['shell_charges'] = shell_charges
        self.results = result['fireball'][-1]

    def get_charges(self, atoms=None):
        if self.results is None:
            try:
                self.read_results()
            except AttributeError:
                return None
        shell_charges = self.results['fireball'][-1]['shell_charges']
        sum_charges = [sum(isc) for isc in shell_charges]
        if atoms is None:
            atoms = self.atoms
        ref_charges = [sum(self.shell_info[s]['occupation']) for s in atoms.symbols]
        charges = [ref-sc for sc, ref in zip(sum_charges, ref_charges)]
        self.results['charges'] = charges
        return charges

    def get_forces(self, atoms=None):
        forces = self.get_property('forces', atoms)
        return np.asarray(forces)

    def get_fermi_level(self):
        return self.get_property('fermi')

    def get_ibz_k_points(self):
        if self.kpt_path is None:
            raise ValueError
        kpts = np.asarray(self.get_kpoints())[:, :3]
        return kpts

    def get_eigenvalues(self, kpt=0, spin=0):
        if self._eigenvalues is not None:
            return self._eigenvalues[spin][kpt]

        if not self.results:
            self.calculate()
        # Read eigen file
        filename = self.sname + '.eigen'
        if not os.path.isfile(filename):
            raise FileExistsError

        with open(filename, 'r') as f:
            lines = f.readlines()

        eigenvalues = []
        eigen = []
        kpts_count = [str(k + 1) for k in range(self.nkpt)]
        for line in lines:
            content = [i.strip() for i in line.strip().split() if len(i.strip()) != 0]
            if len(content) == 0:
                pass
            if len(kpts_count) != 0:
                if content[0] == kpts_count[0]:
                    if len(eigen) != 0:
                        eigenvalues.append(eigen)
                    eigen = []
                    kpts_count.pop(0)
                else:
                    eigen += [float(i) for i in content]
            else:
                eigen += [float(i) for i in content]
        eigenvalues.append(eigen)

        self._eigenvalues = [eigenvalues]  # only one spin

        return self._eigenvalues[spin][kpt]

    def get_number_of_spins(self):
        """
        Current Fireball doesn't support spin polarization.
        :return:
        """
        return 1

    def band_structure(self, atoms=None, reference=None):
        """Create band-structure object for plotting."""
        if atoms is not None:
            self.atoms = atoms.copy()

        from ase.spectrum.band_structure import get_band_structure
        if self.kpt_path is None:
            raise ValueError

        if type(self.kpt_path) in (BandPath,):
            path = self.kpt_path
        else:
            path = None
        return get_band_structure(atoms=self.atoms, calc=self, path=path, reference=reference)

    def read_options(self, input_file='structures.inp', read_atoms=False):
        sname_list = super().read_options(input_file=input_file, read_atoms=read_atoms)
        if not read_atoms or len(sname_list) == 0:
            return None

        if len(sname_list) == 1:
            self.atoms = read_inp(input_file='{}.inp'.format(sname_list[0]))
            kpts_file = '{}.KPOINTS'.format(sname_list[0])
            if os.path.isfile(kpts_file):
                self.read_kpts(kpts_file)
            return None
            # multiple atoms: if it contains multiple atoms, return MultiFireball instead of Fireball
        print("Warning: This file contains multiple atoms, return MultiFireball instead of Fireball")
        atoms_list = []
        for sname in sname_list:
            # read atoms
            atoms = read_inp(input_file='{}.inp'.format(sname))
            # read kpts
            kpts_file = '{}.KPOINTS'.format(sname)
            if os.path.isfile(kpts_file):
                self.read_kpts(kpts_file)
            atoms.set_calculator(self)
            atoms_list.append(atoms)
        multicalc = MultiFireball(atoms_list=atoms_list, sname_list=sname_list, calc=self)
        return multicalc

    @property
    def shell_info(self):
        if self._shell_info is None:
            # read Fdata/info.dat
            self._shell_info = read_info(os.path.join(self.Fdata_path, 'info.dat'))

            # TODO: this is a dirty way to get exicted_label.
            #  The excited label can be obtained by set(ishell). However, still dirty.
            for symbol, value in self._shell_info.items():
                ishell = value['shells']
                len_ground = len(SHELL_PRIMARY_NUMS[value['number']])
                assert len(set(ishell)) <= len_ground  # if not, something wrong during generating Fdata
                excited_label = [0] * len_ground + [1] * (len(ishell) - len_ground)
                self._shell_info[symbol]['excited'] = excited_label
        return self._shell_info

    def get_valence_charge(self, i):
        isymbol = self.atoms.symbols[i]
        shell_info = self.shell_info[isymbol]
        return sum(shell_info['occupation'])

    def write_mwfn(self, kpoint=0):
        mwfn_dict = MWFN_DEFAULT.copy()
        # Initialize default data format
        for k, v in mwfn_dict.items():
            if v is not None:
                mwfn_dict[k] = MWFN_FORMAT[k].format(v)

        # atom information
        mwfn_dict['ncenter'] = MWFN_FORMAT['ncenter'].format(len(self.atoms))
        atoms_coord = [MWFN_FORMAT['atoms_coord'].format(idx + 1,
                                                         iatom.symbol,
                                                         iatom.number,
                                                         self.get_valence_charge(idx),
                                                         *iatom.position)
                       for idx, iatom in enumerate(self.atoms)]
        mwfn_dict['atoms_coord'] = '\n'.join(atoms_coord)

        # cell info
        if np.any(self.atoms.pbc):
            cell_info = {
                'ndim': MWFN_FORMAT['ndim'].format(3),
                'cellv1': (MWFN_FORMAT['cellv1'] * 3).format(*self.atoms.cell[0]),
                'cellv2': (MWFN_FORMAT['cellv2'] * 3).format(*self.atoms.cell[1]),
                'cellv3': (MWFN_FORMAT['cellv3'] * 3).format(*self.atoms.cell[2]),
            }
            mwfn_dict['cell_info'] = CELL_TEMPLATE.substitute(cell_info)
        else:
            mwfn_dict['cell_info'] = ''

        # electron for alpha and beta
        tot_elec = sum([self.get_valence_charge(i) for i in range(len(self.atoms))])
        naelec = 0.5 * tot_elec
        nbelec = 0.5 * tot_elec
        mwfn_dict['naelec'] = MWFN_FORMAT['naelec'].format(naelec)
        mwfn_dict['nbelec'] = MWFN_FORMAT['nbelec'].format(nbelec)
        # total energy
        mwfn_dict['e_tot'] = MWFN_FORMAT['e_tot'].format(self.get_potential_energy() / Hartree)
        # Basis set info
        shell_types = []
        shell_centers = []
        shell_contraction_degress = []
        primitive_exponents = []
        contraction_coefficients = []
        for idx, symbol in enumerate(self.atoms.symbols):
            ishells = self.shell_info[symbol]['shells']
            shell_centers += [idx + 1] * len(ishells)
            shell_types += ishells
            excited_label = self.shell_info[symbol]['excited']
            for shell, is_excited in zip(ishells, excited_label):
                gbs = read_gaussian(self.shell_info[symbol]['number'], shell, is_excited, Fdata_path=self.Fdata_path)
                shell_contraction_degress.append(gbs['degree'])
                primitive_exponents += gbs['alpha'].tolist()
                contraction_coefficients += gbs['coefficient'].tolist()
        nbasis = sum([shell * 2 + 1 for shell in shell_types])
        nshell = len(shell_types)
        nprimshell = len(primitive_exponents)
        nprims = sum([SHELL_PRIMITIVE[st] * degree
                      for st, degree in zip(shell_types, shell_contraction_degress)])
        shell_types = [((i < 2) * 2 - 1) * i for i in shell_types]  # if i >= 2, use negative value

        mwfn_dict['nbasis'] = MWFN_FORMAT['nbasis'].format(nbasis)
        mwfn_dict['nindbasis'] = MWFN_FORMAT['nindbasis'].format(nbasis)
        mwfn_dict['nprims'] = MWFN_FORMAT['nprims'].format(nprims)
        mwfn_dict['nshell'] = MWFN_FORMAT['nshell'].format(nshell)
        mwfn_dict['nprimshell'] = MWFN_FORMAT['nprimshell'].format(nprimshell)

        mwfn_dict['shell_types'] = format_data('shell_types', shell_types)
        mwfn_dict['shell_centers'] = format_data('shell_centers', shell_centers)
        mwfn_dict['shell_contraction_degress'] = format_data('shell_contraction_degress', shell_contraction_degress)
        mwfn_dict['primitive_exponents'] = format_data('primitive_exponents', primitive_exponents)
        mwfn_dict['contraction_coefficients'] = format_data('contraction_coefficients', contraction_coefficients)

        cdcoeffs = read_cdcoeffs(self.sname + '.cdcoeffs-mwfn')[kpoint]
        orbital_coeffs = []
        for orbital in cdcoeffs:
            info = ''.join(orbital['info'])
            coeff = orbital['coeff']
            coeff = reorder_cdcoeffs(coeff, shell_types)
            orbital_coeffs += [info + format_data('MO_coefficients', coeff)]
        mwfn_dict['orbital_coeffs'] = '\n\n'.join(orbital_coeffs)

        self.mwfn_dict = mwfn_dict
        content = MWFN_TEMPLATE.substitute(mwfn_dict)
        # TODO: read orbital info, append to the content

        # write out
        with open(self.sname + '.mwfn', 'w') as f:
            f.write(content)


class MultiFireball:
    name = 'multi_fireball'

    def __init__(self, atoms_list=None, calc=None, sname_list=None):
        if calc is not None:
            self.calc = calc
        elif atoms_list is not None:
            if atoms_list[0].calc is not None:
                self.calc = atoms_list[0].calc
        else:
            self.calc = None

        _ = [atoms.set_calc(self.calc) for atoms in atoms_list if atoms.calc is None]

        self.atoms_list = atoms_list

        if sname_list is None:
            self.sname_list = ["{:03d}".format(idx + 1) for idx in range(len(self.atoms_list))]

    def write_Fdata_inp(self, Fdata_path=None):
        if Fdata_path is None:
            Fdata_path = 'Fdata'
        species = sorted(list(set([i for atoms in self.atoms_list for i in set(atoms.numbers)])))
        with open('Fdata.inp', 'w') as f:
            f.write("{}\n".format(len(species)))
            for s in species:
                f.write("{}\n".format(s))
            f.write("'{}'".format(Fdata_path))

    def write_options(self):
        with open('structures.inp', 'w') as f:
            f.write("{:d}\n".format(len(self.atoms_list)))
            for sname in self.sname_list:
                f.write("{}.inp\n".format(sname))

            f.write("! Write out options - leave this line as comment line or null\n")

            f.write("&OUTPUT\n")
            write_params(self.calc.output_params, f)
            f.write("&END\n")

            f.write("&OPTIONS\n")
            write_params(self.calc.options_params, f)
            f.write("&END\n")

            f.write("&XSFOPTIONS\n")
            write_params(self.calc.xsfoptions_params, f)
            f.write("&END\n")

    def write_input(self):
        Fdata_path = self.calc.Fdata_path
        self.write_Fdata_inp(Fdata_path=Fdata_path)
        self.write_options()
        for atoms, sname in zip(self.atoms_list, self.sname_list):
            atoms.calc.sname = sname
            atoms.calc.write_atoms(atoms=atoms, pbc=atoms.pbc)
            atoms.calc.write_kpts()

    def calculate(self):
        errorcode = self._run(command=self.calc.command,
                              directory=self.calc.directory)
        if errorcode:
            raise CalculationFailed(
                '{} in {} returned an error: {:d}'.format(
                    self.name, self.calc.directory, errorcode))
        for atoms in self.atoms_list:
            atoms.calc.read_results()

    def _run(self, command=None, directory=None):
        """Method to explicitly execute Fireball"""
        if command is None:
            command = self.calc.command
        if directory is None:
            directory = self.calc.directory
        errorcode = subprocess.call(command, shell=True, cwd=directory)
        return errorcode


def read_inp(input_file):
    # read fireball input file, return ase.Atoms
    with open(input_file, 'r') as f:
        lines = f.readlines()

    natoms, is_cluster = lines[0].split()
    pbc = not bool(int(is_cluster))
    cell = [[float(i) for i in lines[iline + 1].split()] for iline in range(3)]
    numbers = []
    positions = []
    for line in lines[4:]:
        content = line.strip().split()
        if len(content) == 4:
            num, x, y, z = content
            numbers.append(int(num))
            positions.append(list(map(float, [x, y, z])))

    atoms = ase.Atoms(numbers=numbers,
                      positions=positions,
                      cell=cell,
                      pbc=pbc)
    return atoms



from fireball_calculator.fireball import Fireball
import ase
from ase.io.trajectory import Trajectory
import numpy as np
from ase.io import read
from ase.units import kJ
from ase.eos import EquationOfState


cell = np.array([[2.715000, 2.715000, 0.000000],
                 [2.715000, 0.000000, 2.715000],
                 [0.000000, 2.715000, 2.715000]])
positions = np.array([[0.0000000, 0.0000000, 0.0000000],
                      [1.3575000, 1.3575000, 1.3575000]])
kwargs = {'kpt_size': [2, 2, 2],
          'iwriteout_ME_SandH': 0,
          'iwriteout_density': 0,
          'iwriteout_cdcoeffs': 0,
          'iwriteout_charges': 1,
          'iwriteout_energies': 0,
          'iwriteout_populations': 0,
          'iwriteout_forces': 1,
          'iwriteout_neighbors': 0,
          'iwriteout_dos': 0,
          'iwriteout_abs': 0,
          'iwriteout_ewf': 0,
          'nstepi': 1,
          'nstepf': 100,
          'iquench': -3,
          'T_initial': 0.0,
          'T_final': 0.0,
          'T_want': 0.0,
          'taurelax': 5.0,
          'efermi_T': 200.0,
          'dt': 1.00,
          'iensemble': 0,
          'iconstraint_rcm': 1,
          'iconstraint_vcm': 1,
          'iconstraint_L': 0,
          'iconstraint_KE': 1,
          'ifix_neighbors': 0,
          'ifix_CHARGES': 0,
          'max_scf_iterations_set': 100,
          'scf_tolerance_set': 0.00000001,
          'beta_set': 0.04,
          'Ecut_set': 200.0,
          'rho_surface_min': 0.0005,
          'rho_surface_max': 0.01000000,
          }

cell_factors = np.linspace(0.8, 1.2, 8)
traj = Trajectory('Si.traj', 'w')
for cf in cell_factors:
    atoms = ase.Atoms(numbers=[14, 14],
                      cell=cell,
                      pbc=True,
                      positions=positions,
                      )
    atoms.set_cell(cell=cell*cf, scale_atoms=True)
    calc = Fireball(command='/home/ren/bin/lightning.3.x', **kwargs)
    atoms.set_calculator(calc)
    e0 = atoms.get_potential_energy()
    traj.write(atoms)
    print("The energy for cell factor {:.3f} is {:.3f}".format(cf, e0))

configs = read('Si.traj', index=':')
volumes = [si.get_volume() for si in configs]
energies = [si.get_potential_energy() for si in configs]
eos = EquationOfState(volumes, energies)
v0, e0, B = eos.fit()
print(B / kJ * 1.0e24, 'GPa')
eos.plot('Si-eos.png')

print("Done!!")
